#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn

def delta_dirac_dist(memory, samples,
    k=10, kernel_cluster_distance=0.008, kernel_epsilon=0.0001,
    c=0.001, sm=8, angular_space=False):
    if angular_space:
      dist_mat = torch.abs(memory[:,None,:] - samples[None,:,:]);
      cond = dist_mat > 0.5
      dist_mat[cond] = 1 - dist_mat[cond]
      dist_mat = torch.sum(dist_mat, dim=-1)
      # print(torch.max(dist_mat), torch.min(dist_mat))
    else:
      dist_mat = torch.cdist(memory, samples, p=2)
    # # TODO: moving average
    # dist = dist / np.mean(dist)

    # dist_mat, _ = torch.max(dist_mat - kernel_cluster_distance, 0)
    # print(dist_mat.shape)
    # TODO: Test this kernel
    # kernel = 1 / (dist + kernel_epsilon)
    kernel = kernel_epsilon / (dist_mat + kernel_epsilon)
    s = torch.sqrt(torch.sum(kernel, dim=0)) + c
    reward = 1/s
    index = torch.argmax(reward, axis=0)
    # if np.isnan(s) or s > sm:
    #     return 0
    return index, dist_mat

def dirac_sampling(memory, samples):
    scale_ind,_ = delta_dirac_dist(memory[:,0:3], samples[:,0:3], angular_space=False)
    trans_ind,_ = delta_dirac_dist(memory[:,3:6], samples[:,3:6], angular_space=False)
    angle_ind,_ = delta_dirac_dist(memory[:,6:9], samples[:,6:9], angular_space=True)
    chosen_aug = torch.hstack((samples[[scale_ind],0:3], samples[[trans_ind],3:6], samples[[angle_ind],6:9]))
    rl_inds = [scale_ind, trans_ind, angle_ind]
    return rl_inds, chosen_aug


def cosine_distance(z1, z2, association, cosine_distance=False):
		#z1,z2: [bs, feat_dim, no_points]; intermediate features
		#association: [bs, no_points]; maps z1 to z2 indices for loss
		feat_dim = z1.shape[1]
		association2 = torch.unsqueeze(association,1)
		association2 = association2.repeat(1,feat_dim,1)
		z1_2_g = torch.gather(z1, 2, association2) #transformed from 1 to 2 using associations
		###--- this block is to verify if gather is workign correct ---
		# z1_2 =[]
		# batch = np.arange(z1.shape[0])
		# for i in batch:
		# 	z1_2.append(z1[i, :, association[i,:]])
		# z1_2 = torch.stack(z1_2, dim=0)
		# print("->", torch.all(torch.eq(z1_2, z1_2_g)), z1_2[0,0,0], z1_2_g[0,0,0])
		###--------------------------------
		if cosine_distance:
			cos = nn.CosineSimilarity(dim=1, eps=1e-6)
			cos_sim = cos(z1_2_g, z2)
			cos_dist = torch.ones_like(cos_sim)-cos_sim
			loss = torch.sum(cos_dist)
		else:
			mse = nn.MSELoss(reduction="none")
			print(torch.var_mean(z1_2_g,unbiased=True), torch.var_mean(z2, unbiased=True))
			loss = mse(z1_2_g, z2)
			loss = torch.mean(loss, dim=1)
			loss = torch.sum(loss)
		return loss

class RMSELoss(nn.Module):
		def __init__(self, reduction="mean", eps=1e-8):
				super().__init__()
				self.reduction=reduction
				self.mse = nn.MSELoss(reduction=self.reduction)
				self.eps = eps

		def forward(self,yhat,y):
				if self.reduction=="none":
					loss = self.mse(yhat,y)
					loss = torch.sqrt(torch.mean(loss, dim=-1) + self.eps)
				else:
					loss = torch.sqrt(self.mse(yhat,y) + self.eps)
				return loss

def cartesian(arrays, out=None):
		"""
		Generate a cartesian product of input arrays.

		Parameters
		----------
		arrays : list of array-like
				1-D arrays to form the cartesian product of.
		out : ndarray
				Array to place the cartesian product in.

		Returns
		-------
		out : ndarray
				2-D array of shape (M, len(arrays)) containing cartesian products
				formed of input arrays.

		Examples
		--------
		>>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
		array([[1, 4, 6],
					 [1, 4, 7],
					 [1, 5, 6],
					 [1, 5, 7],
					 [2, 4, 6],
					 [2, 4, 7],
					 [2, 5, 6],
					 [2, 5, 7],
					 [3, 4, 6],
					 [3, 4, 7],
					 [3, 5, 6],
					 [3, 5, 7]])

		"""

		arrays = [np.asarray(x) for x in arrays]
		dtype = arrays[0].dtype

		n = np.prod([x.size for x in arrays])
		if out is None:
				out = np.zeros([n, len(arrays)], dtype=dtype)

		#m = n / arrays[0].size
		m = int(n / arrays[0].size)
		out[:,0] = np.repeat(arrays[0], m)
		if arrays[1:]:
				cartesian(arrays[1:], out=out[0:m, 1:])
				for j in range(1, arrays[0].size):
				#for j in xrange(1, arrays[0].size):
						out[j*m:(j+1)*m, 1:] = out[0:m, 1:]
		return out

def cal_loss(pred, gold, smoothing=True):
		''' Calculate cross entropy loss, apply label smoothing if needed. '''

		gold = gold.contiguous().view(-1)

		if smoothing:
				eps = 0.2
				n_class = pred.size(1)

				one_hot = torch.zeros_like(pred).scatter(1, gold.view(-1, 1), 1)
				one_hot = one_hot * (1 - eps) + (1 - one_hot) * eps / (n_class - 1)
				log_prb = F.log_softmax(pred, dim=1)

				loss = -(one_hot * log_prb).sum(dim=1).mean()
		else:
				loss = F.cross_entropy(pred, gold, reduction='mean')

		return loss


class IOStream():
		def __init__(self, path):
				self.f = open(path, 'a')

		def cprint(self, text):
				print(text)
				self.f.write(text+'\n')
				self.f.flush()

		def close(self):
				self.f.close()
