#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
contains both supervised and unsupervised training for two different datasets.
model classification and pointcloud segmentation
"""


from __future__ import print_function
import os
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.optim.lr_scheduler import CosineAnnealingLR
import numpy as np
from torch.utils.data import DataLoader
import sklearn.metrics as metrics
from tqdm import tqdm
from config.config import *
import sklearn.metrics as metrics
from lib.visualize import viz_feats
from lib.model import PointNet, DGCNN, DGCNN_partseg
from lib.util import IOStream, RMSELoss
from lib.dataloaders import Data3DUnsupervised, Data3D
from lightly.loss.ntx_ent_loss import NTXentLoss
from train_svm import train_val_gen, train_svm
import copy
import joblib
from scipy.spatial import distance

# load args
args = parse_args()

device = torch.device("cuda" if args.cuda else "cpu")

criterion = NTXentLoss(temperature = 0.1).to(device)

### setup dataloader
dataset = Data3DUnsupervised
dataset_svm = Data3D
### setup model
if args.model == 'pointnet':
		model = PointNet(args).to(device)
elif args.model == 'dgcnn':
		if args.task =="classification":
			model = DGCNN(args).to(device)
		if args.task =='segmentation':
			model = DGCNN_partseg(args).to(device)
else:
		raise Exception("Not implemented")

# if args.rl_based_search:
# 	rl_model = ValueNet(args).to(device)
# 	rmse_loss = RMSELoss()
# else:
# 	rl_model = None

def _init_(args):
		# copying models and files to corresponding checkpoints
		if not os.path.exists('checkpoints'):
				os.makedirs('checkpoints')
		if not os.path.exists('checkpoints/'+args.exp_name):
				os.makedirs('checkpoints/'+args.exp_name)
		if not os.path.exists('checkpoints/'+args.exp_name+'/'+'models'):
				os.makedirs('checkpoints/'+args.exp_name+'/'+'models')
		if not os.path.exists(args.output_folder):
				os.makedirs(args.output_folder)

def train_val_unsup(data_loader, dataset, model, epoch=0, opt=None, phase="test", viz_feats=False):
		# train-val method for unsupervised approach
		# unsupervised training needs different data_loader, losses and metrics
		train_corr_loss = 0.0
		train_rl_loss = 0.0
		count = 0.0

		if phase=="train":
			model.train()

		dataloader_iterator = iter(data_loader)
		iterations = len(data_loader)
		dataset.reset_mem()
		for indx in tqdm(range(iterations)):
				dataset.reset_augs()
				max_dist = 0
				try:
						data1, data2, association = next(dataloader_iterator)
				except StopIteration:
						break

				data1, data2, association = data1.to(device), data2.to(device).squeeze(), association.to(device)
				data1 = data1.permute(0, 2, 1)#bs, ndim, no of points
				data2 = data2.permute(0, 2, 1)
				# print(data1.shape)
				batch_size = data1.size()[0]
				opt.zero_grad()
				if args.w_association:
					midfeats1, feats1, _ = model(data1) # full pointcloud data
					midfeats2, feats2, _ = model(data2) # cropped pointcloud data
					_, feats3, _ = model(data1, association=association) # full pointcloud data, but with feature association before projection
					if args.association_type==1: #loss without any projection of associated feats
						#converting to same shape as feats
						association2 = torch.unsqueeze(association,1)
						association2 = association2.repeat(1, midfeats1.shape[1],1)
						midfeats1_a = torch.gather(midfeats1, 2, association2)
						midfeats1_a = midfeats1_a.view(midfeats1_a.shape[0], -1)
						midfeats2 = midfeats2.view(midfeats2.shape[0], -1)
						loss = criterion(midfeats1_a, midfeats2) + criterion(feats1, feats2)
					if args.association_type==2: #loss with projection of associated feats. 1-way
						loss = criterion(feats3, feats2)
					if args.association_type==3: #loss with projection of associated feats. 2-way
						loss = criterion(feats1, feats2)+criterion(feats3, feats2)
					if args.association_type==4: #loss with projection of associated feats. 3-way
						loss = criterion(feats1, feats2)+criterion(feats1, feats3)+criterion(feats3, feats2)
				else:
					_, feats1, _ = model(data1)
					_, feats2, _ = model(data2)
					loss = criterion(feats1, feats2)
				if phase=="train":
					loss.backward()
					opt.step()
				##visualize
				if indx%5000==0 and viz_feats:
					viz_feats(data1[0,...].cpu().numpy().T, feats1[0,...].detach().cpu().numpy().T, phase+"_"+str(indx).zfill(4)+"_"+str(epoch).zfill(3))

				count += batch_size
				train_corr_loss += loss.item() * batch_size
		# evaluation metrics
		lr = opt.param_groups[0]['lr']

		value_str = '%d, %.6f, %.6f, %.6f\n'% (epoch, lr,
																			 train_corr_loss*1.0/count, max_dist)

		outstr = '%s %d, cross correlation loss: %.6f' % (phase, epoch,
																										 train_corr_loss*1.0/count)
		with open(args.output_folder+'/'+args.exp_name+'_'+phase+'.txt','a+') as file:
			file.write(value_str)

		io.cprint(outstr)
		return train_corr_loss

def train(args, io):
		# train method abstraction for both supervised and unsupervised

		global dataset, model

		all_params = []
		# if args.rl_based_search:
		# 	rl_model = nn.DataParallel(rl_model)
		# 	all_params+=rl_model.parameters()
		print(str(model))

		train_dataset = dataset(partition='train', args=args)
		train_loader = DataLoader(train_dataset, num_workers=0,
															batch_size=args.batch_size, shuffle=True, drop_last=True)

		val_dataset = dataset(partition='test', args=args)
		val_loader = DataLoader(val_dataset, num_workers=0,
															batch_size=args.batch_size, shuffle=False, drop_last=True)

		## svm dataset
		args_svm = copy.deepcopy(args)
		args_svm.dataset = args.svm_dataset
		args_svm.w_aug = False
		train_dataset_svm = dataset_svm(partition='train', args=args_svm)
		train_loader_svm = DataLoader(train_dataset_svm, num_workers=0,
															batch_size=args.batch_size, shuffle=False, drop_last=True)
		val_dataset_svm = dataset_svm(partition='test', args=args_svm)
		val_loader_svm = DataLoader(val_dataset_svm, num_workers=0,
															batch_size=args.batch_size, shuffle=False, drop_last=True)

		print(str(model))
		model = nn.DataParallel(model)
		if args.load_pretrained:
			model.load_state_dict(torch.load(args.load_model_path))

		all_params+=model.parameters()

		print("Let's use", torch.cuda.device_count(), "GPUs!")

		if args.use_sgd:
				print("Use SGD")
				opt = optim.SGD(all_params, lr=args.lr*100, momentum=args.momentum, weight_decay=1e-4)
		else:
				print("Use Adam")
				opt = optim.Adam(all_params, lr=args.lr, weight_decay=1e-4)

		# according to dgcnn model for sgd
		scheduler = CosineAnnealingLR(opt, args.epochs, eta_min=args.lr)

		best_val_score = 0
		# best_rl_loss = float('inf')
		for epoch in range(args.epochs):
			scheduler.step()

			train_loss = train_val_unsup(train_loader, train_dataset, model, epoch, opt, phase="train", viz_feats=args.viz_feats)

			val_loss = train_val_unsup(val_loader, val_dataset, model, epoch, opt, phase="val", viz_feats=args.viz_feats)

			## validate with svm
			train_feats, train_labels = train_val_gen(train_loader_svm, model, epoch, opt, phase="train", viz_feats=args.viz_feats)
			print(train_feats.shape, train_labels.shape)
			val_feats, val_labels = train_val_gen(val_loader_svm, model, epoch, opt, phase="val", viz_feats=args.viz_feats)
			score, svm_model = train_svm(train_feats, train_labels, val_feats, val_labels)
			print("svm accuracy:", score)

			lr = opt.param_groups[0]['lr']
			value_str = '%d, %.6f, %.6f\n'% (epoch, lr, score)
			with open(args.output_folder+'/'+args.exp_name+'_val_svm.txt','a+') as file:
				file.write(value_str)

			if score > best_val_score:
					best_val_score = score
					torch.save(model.state_dict(), 'checkpoints/%s/models/model.t7' % args.exp_name)
					# save the model to disk
					svm_file = 'checkpoints/%s/models/model_svm.sav' % args.exp_name
					joblib.dump(svm_model, svm_file)

			torch.save(model.state_dict(), 'checkpoints/%s/models/model_last.t7' % args.exp_name)
			svm_file = 'checkpoints/%s/models/model_last_svm.sav' % args.exp_name
			joblib.dump(svm_model, svm_file)
			# if train_rl_loss < best_rl_loss and args.rl_based_search:
			# 		best_rl_loss = train_rl_loss
			# 		torch.save(rl_model.state_dict(), 'checkpoints/%s/models/rl_model.t7' % args.exp_name)

if __name__ == "__main__":

		_init_(args)

		io = IOStream('checkpoints/' + args.exp_name + '/run.log')
		io.cprint(str(args))

		with open(args.output_folder+'/'+args.exp_name+'_train.txt','w+') as f:
			f.write("epoch, learning rate, correlation loss, rl loss\n")

		with open(args.output_folder+'/'+args.exp_name+'_val.txt','w+') as f:
			f.write("epoch, learning rate, correlation loss, rl loss\n")

		with open(args.output_folder+'/'+args.exp_name+'_val_svm.txt','w+') as f:
			f.write("epoch, learning rate, loss\n")


		args.cuda = args.cuda and torch.cuda.is_available()
		torch.manual_seed(args.seed)
		if args.cuda:
				io.cprint(
						'Using GPU : ' + str(torch.cuda.current_device()) + ' from ' + str(torch.cuda.device_count()) + ' devices')
				torch.cuda.manual_seed(args.seed)
		else:
				io.cprint('Using CPU')

		if not args.eval:
				train(args, io)
		else:
				test(args, io)
